/**
 * @author Titus Wormer
 * @copyright 2015 Titus Wormer. All rights reserved.
 * @module File
 * @fileoverview Virtual file format to attach additional
 *   information related to the processed input.  Similar
 *   to`wearefractal/vinyl`.  Additionally, File can be
 *   passed directly to an ESLint formatter to visualise
 *   warnings and errors relating to a file.
 */

'use strict';

/**
 * ESLint's formatter API expects `filePath` to be a
 * string.  This hack supports invocation as well as
 * implicit coercion.
 *
 * @example
 *   var file = new File();
 *   filePath = filePathFactory(file);
 *
 * @param {File} file
 * @return {Function}
 */
function filePathFactory(file) {
    /**
     * Get the location of `file`.
     *
     * Returns empty string when without `filename`.
     *
     * @example
     *   var file = new File({
     *     'directory': '~',
     *     'filename': 'example',
     *     'extension': 'markdown'
     *   });
     *
     *   String(file.filePath); // ~/example.markdown
     *   file.filePath() // ~/example.markdown
     *
     * @property {Function} toString - Itself.
     * @return {string}
     */
    function filePath() {
        var directory;

        if (file.filename) {
            directory = file.directory;

            if (directory.charAt(directory.length - 1) === '/') {
                directory = directory.slice(0, -1);
            }

            if (directory === '.') {
                directory = '';
            }

            return (directory ? directory + '/' : '') +
                file.filename +
                (file.extension ? '.' + file.extension : '');
        }

        return '';
    }

    filePath.toString = filePath;

    return filePath;
}

/**
 * Construct a new file.
 *
 * @example
 *   var file = new File({
 *     'directory': '~',
 *     'filename': 'example',
 *     'extension': 'markdown',
 *     'contents': 'Foo *bar* baz'
 *   });
 *
 *   file === File(file) // true
 *   file === new File(file) // true
 *   File('foo') instanceof File // true
 *
 * @constructor
 * @class {File}
 * @param {Object|File|string} [options] - either an
 *   options object, or the value of `contents` (both
 *   optional).  When a `file` is passed in, it's
 *   immediately returned.
 */
function File(options) {
    var self = this;

    if (!(self instanceof File)) {
        return new File(options);
    }

    if (options instanceof File) {
        return options;
    }

    if (!options) {
        options = {};
    } else if (typeof options === 'string') {
        options = {
            'contents': options
        };
    }

    self.filename = options.filename || null;
    self.contents = options.contents || '';

    self.directory = options.directory === undefined ? '' : options.directory;

    self.extension = options.extension === undefined ?
        'md' : options.extension;

    self.messages = [];

    /*
     * Make sure eslint’s formatters stringify `filePath`
     * properly.
     */

    self.filePath = filePathFactory(self);
}

/**
 * Move a file by passing a new directory, filename,
 * and extension.  When these are not given, the default
 * values are kept.
 *
 * @example
 *   var file = new File({
 *     'directory': '~',
 *     'filename': 'example',
 *     'extension': 'markdown',
 *     'contents': 'Foo *bar* baz'
 *   });
 *
 *   file.move({'directory': '/var/www'});
 *   file.filePath(); // '/var/www/example.markdown'
 *
 *   file.move({'extension': 'md'});
 *   file.filePath(); // '/var/www/example.md'
 *
 * @this {File}
 * @param {Object} options
 */
function move(options) {
    var self = this;

    if (!options) {
        options = {};
    }

    self.directory = options.directory || self.directory || '';
    self.filename = options.filename || self.filename || null;
    self.extension = options.extension || self.extension || 'md';
}

/**
 * Stringify a position.
 *
 * @example
 *   stringify({'line': 1, 'column': 3}) // '1:3'
 *   stringify({'line': 1}) // '1:1'
 *   stringify({'column': 3}) // '1:3'
 *   stringify() // '1:1'
 *
 * @param {Object?} [position] - Single position, like
 *   those available at `node.position.start`.
 * @return {string}
 */
function stringify(position) {
    if (!position) {
        position = {};
    }

    return (position.line || 1) + ':' + (position.column || 1);
}

/**
 * Warn.
 *
 * Creates an exception (see `File#exception()`),
 * sets `fatal: false`, and adds it to the file's
 * `messages` list.
 *
 * @example
 *   var file = new File();
 *   file.warn('Something went wrong');
 *
 * @this {File}
 * @param {string|Error} reason - Reason for warning.
 * @param {Node|Location|Position} [position] - Location
 *   of warning in file.
 * @return {Error}
 */
function warn(reason, position) {
    var err = this.exception(reason, position);

    err.fatal = false;

    this.messages.push(err);

    return err;
}

/**
 * Fail.
 *
 * Creates an exception (see `File#exception()`),
 * sets `fatal: true`, adds it to the file's
 * `messages` list.  If `quiet` is not true,
 * throws the error.
 *
 * @example
 *   var file = new File();
 *   file.fail('Something went wrong'); // throws
 *
 * @this {File}
 * @throws {Error} - When not `quiet: true`.
 * @param {string|Error} reason - Reason for failure.
 * @param {Node|Location|Position} [position] - Location
 *   of failure in file.
 * @return {Error} - Unless thrown, of course.
 */
function fail(reason, position) {
    var err = this.exception(reason, position);

    err.fatal = true;

    this.messages.push(err);

    if (!this.quiet) {
        throw err;
    }

    return err;
}

/**
 * Create a pretty exception with `reason` at `position`.
 * When an error is passed in as `reason`, copies the
 * stack.  This does not add a message to `messages`.
 *
 * @example
 *   var file = new File();
 *   var err = file.exception('Something went wrong');
 *
 * @this {File}
 * @param {string|Error} reason - Reason for message.
 * @param {Node|Location|Position} [position] - Location
 *   of message in file.
 * @return {Error} - An object including file information,
 *   line and column indices.
 */
function exception(reason, position) {
    var file = this.filePath();
    var message = reason.message || reason;
    var location;
    var err;

    /*
     * Node / location / position.
     */

    if (position && position.position) {
        position = position.position;
    }

    if (position && position.start) {
        location = stringify(position.start) + '-' + stringify(position.end);
        position = position.start;
    } else {
        location = stringify(position);
    }

    err = new Error(message);

    err.name = (file ? file + ':' : '') + location;
    err.file = file;
    err.reason = message;
    err.line = position ? position.line : null;
    err.column = position ? position.column : null;

    if (reason.stack) {
        err.stack = reason.stack;
    }

    return err;
}

/**
 * Check if `file` has a fatal message.
 *
 * @example
 *   var file = new File();
 *   file.quiet = true;
 *   file.hasFailed; // false
 *
 *   file.fail('Something went wrong');
 *   file.hasFailed; // true
 *
 * @this {File}
 * @return {boolean}
 */
function hasFailed() {
    var messages = this.messages;
    var index = -1;
    var length = messages.length;

    while (++index < length) {
        if (messages[index].fatal) {
            return true;
        }
    }

    return false;
}

/**
 * Create a string representation of `file`.
 *
 * @example
 *   var file = new File('Foo');
 *   String(file); // 'Foo'
 *
 * @this {File}
 * @return {string} - value at the `contents` property
 *   in context.
 */
function toString() {
    return this.contents;
}

/*
 * Methods.
 */

File.prototype.move = move;
File.prototype.exception = exception;
File.prototype.toString = toString;
File.prototype.warn = warn;
File.prototype.fail = fail;
File.prototype.hasFailed = hasFailed;

/*
 * Expose.
 */

module.exports = File;
